#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os, sys, platform, uuid, re, shutil, tempfile
# import cssutils, logging 
from os.path import expanduser, basename
from subprocess import Popen, PIPE
from io import BytesIO
from datetime import datetime
from PIL import Image

from plugin_utils import Qt, QtCore, QtGui, QtWidgets
from plugin_utils import loadUi, PluginApplication, iswindows

# required for loading the .ui and .qm files from the plugin folder
plugin_dir = os.path.dirname(os.path.abspath(__file__))

# auxiliary KindleUnpack libraries for adding ASINs and azw3/mobi splitting
from dualmetafix_mmap import DualMobiMetaFix, pathof, iswindows
from mobi_split import mobi_split

# auxiliary tools for metadata and epub handling
from epub_utils import epub_zip_up_book_contents
from quickparser import QuickXHTMLParser
from compatibility_utils import unquoteurl

# detect OS
isosx = sys.platform.startswith('darwin')
islinux = sys.platform.startswith('linux')

#---------------------------------------------------------------------------
# credit: DiapDealer 
# https://www.mobileread.com/forums/showpost.php?p=3575920&postcount=229
#---------------------------------------------------------------------------
def find_python3lib_dir(appdir):
    ''' returns the folder with the python libraries bundled with Sigil '''
    extralibdir = None
    # Determine python3lib path per platform
    if sys.platform.startswith('darwin'):
        extralibdir = os.path.join(appdir, '../python3lib')
    elif sys.platform.startswith('win'):
        extralibdir = os.path.join(appdir, 'python3lib')
    else:
        for path in sys.path:
            if 'plugin_launchers/python' in path:
                extralibdir = os.path.join(path, '../../python3lib')
                break
    return extralibdir

#--------------------------------------------------------------------------
#decode xml encoded strings for older versions < 20190716
#--------------------------------------------------------------------------
def xmldecode(data):
    ''' replaces entities with characters '''
    if data is None:
        return ''
    newdata = data
    newdata = newdata.replace('&quot;', '"')
    newdata = newdata.replace('&gt;', '>')
    newdata = newdata.replace('&lt;', '<')
    newdata = newdata.replace('&amp;', '&')
    return newdata

#----------------------------------------------------------------------
# bk.getnavid() for older versions < 20190716
#----------------------------------------------------------------------
def getnavid(bk):
    ''' returns the manifest id of the navigation document '''
    if bk.epub_version == "2.0":
        return None
    for id in sorted(bk._w.id_to_mime):
        mime = bk._w.id_to_mime[id]
        if mime == "application/xhtml+xml":
            properties = bk._w.id_to_props[id]
            if properties is not None and "nav" in properties:
                return id
    return None

#------------------------------------------------------------------
# returns landmarks for older versions < 20190716
#------------------------------------------------------------------
def getLandmarks(navdata):
    ''' returns ordered list of tuples (epubtype, href, title)
    href is unquoted (percent encodings removed)
    title has been xml decoded/unescaped '''
    landmarks = []
    
    qp = QuickXHTMLParser()
    qp.setContent(navdata)
    title = ""
    nav_type = None
    href = None
    epubtype = None
    for txt, tp, tname, ttype, tattr in qp.parse_iter():
        if txt is not None:
            if ".a." in tp or tp.endswith(".a"):
                title = title + txt
            else:
                title = ""
        else:
            if tname == "nav" and ttype == "begin":
                nav_type = tattr.get("epub:type", None)
                continue
            if tname == "nav" and ttype == "end":
                nav_type = None
                continue

            if nav_type is not None and nav_type == "landmarks": 
                if tname == "a" and ttype == "begin":
                    href = tattr.get("href", "")
                    href = unquoteurl(href)
                    epubtype = tattr.get("epub:type", None)
                    continue
                if tname == "a" and ttype == "end":
                    if epubtype is not None:
                        title = xmldecode(title)
                        landmarks.append((epubtype, href, title))
                    title = ""
                    epubtype = None
                    href=None
                    continue
    return landmarks

#----------------------------------------------------------
# get 'C:\Users\<User>\AppData\Local\' folder location
#----------------------------------------------------------
def GetLocalAppData():
    ''' returns the local AppData folder '''
    # check for Windows 7 or higher
    if sys.getwindowsversion().major >= 6:
        return os.path.join(os.getenv('USERPROFILE'), 'AppData', 'Local')
    else:
        return os.path.join(os.getenv('USERPROFILE'), 'Local Settings', 'Application Data')

#------------------------
# get desktop folder
#------------------------
def GetDesktop():
    ''' returns the desktop/home folder '''
    # output directory
    home = expanduser('~')
    desktop = os.path.join(home, 'Desktop')
    if os.path.isdir(desktop):
        return desktop
    else:
        # use home folder for OSes with localized Desktop names
        return home

#--------------------------------
# download kindlegen for Linux
#--------------------------------
def download_linux_kg(bk):
    ''' downloads kindlegen from archive.org '''
    linux_kindlegen_url = 'https://web.archive.org/web/20150803131026if_/https://kindlegen.s3.amazonaws.com/kindlegen_linux_2.6_i386_v2_9.tar.gz'
    kg_path = None
    from urllib.request import urlretrieve
    import tarfile
    tar_gz_file_name = os.path.join(bk._w.plugin_dir, bk._w.plugin_name, 'kindlegen_linux_2.6_i386_v2_9.tar.gz')
    out_dir = os.path.join(expanduser('~'), 'bin')
    
    if os.path.isfile(tar_gz_file_name): os.remove(tar_gz_file_name) 
    print('Downloading kindlegen_linux_2.6_i386_v2_9.tar.gz...')
    try:
        urlretrieve(linux_kindlegen_url, tar_gz_file_name)
    except:
        pass

    if os.path.isfile(tar_gz_file_name):
        print('kindlegen_linux_2.6_i386_v2_9.tar.gz downloaded.')
        tar = tarfile.open(tar_gz_file_name, "r:gz")
        for member in tar.getmembers():
            if member.name.endswith('kindlegen'):
                tar.extract(member, out_dir)
                kg_bin_path = os.path.join(out_dir, 'kindlegen')
                if os.path.isfile(kg_bin_path):
                    print('kindlegen extracted to {}.'.format(kg_bin_path))
                    kg_path = kg_bin_path
                else:
                    print('kindlegen could not be extracted to {}.'.format(kg_bin_path))
                break
        else:
            print('kindlegen_linux_2.6_i386_v2_9.tar.gz couldn\'t be downloaded.')

    return kg_path

#----------------------------
# get kindlegen path
#----------------------------
def findKindleGen(bk):
    ''' returns the KindleGen path '''
    kg_path = None
    calibre_debug_path = None

    #-----------------------
    # Windows
    #-----------------------

    if iswindows:
        # C:\Users\<User>\AppData\Local\Amazon\Kindle Previewer 3\lib\fc\bin\kindlegen.exe
        default_windows_path = os.path.join(GetLocalAppData(), 'Amazon', 'Kindle Previewer 3', 'lib', 'fc', 'bin', 'kindlegen.exe')
        if os.path.isfile(default_windows_path):
            kg_path = default_windows_path
        else:
            which_windows_path = shutil.which('kindlegen.exe')
            if which_windows_path: kg_path = which_windows_path
            
        # "C:\Program Files\Calibre2\calibre-debug.exe"
        default_calibre_debug_path = os.path.join(os.environ["ProgramW6432"], 'Calibre2', 'calibre-debug.exe')
        if os.path.isfile(default_calibre_debug_path):
            calibre_debug_path = default_calibre_debug_path
        else:
            windows_calibre_debug_path = shutil.which('calibre-debug.exe')
            if windows_calibre_debug_path: calibre_debug_path = windows_calibre_debug_path

    #----------------------
    # Linux
    #---------------------

    if islinux:
        default_linux_path = os.path.join('/usr', 'local', 'bin', 'kindlegen')
        default_linux_path2 = os.path.join(expanduser('~'), 'bin', 'kindlegen')

        if os.path.isfile(default_linux_path):
            kg_path = default_linux_path
        elif os.path.isfile(default_linux_path2):
            kg_path = default_linux_path2
        else:
            which_linux_path = shutil.which('kindlegen')
            if which_linux_path: kg_path = which_linux_path

        if not kg_path:
            answer = ok_msg_box('kindlegen not found.', 'Do you want to download a copy from archive.org?', bk)
            if answer:
                kg_path = download_linux_kg(bk)

        # kfx output probably won't work on Linux...
        linux_calibre_debug_path = shutil.which('calibre-debug')
        if linux_calibre_debug_path: calibre_debug_path = linux_calibre_debug_path

    #----------------------------------
    # macOS
    #----------------------------------

    if isosx:
        # /Applications/Kindle Previewer 3.app/Contents/lib/fc/bin/kindlegen
        default_osx_path = os.path.join('/Applications', 'Kindle Previewer 3.app', 'Contents', 'lib', 'fc', 'bin', 'kindlegen')

        if os.path.isfile(default_osx_path):
            kg_path = default_osx_path
        else:
            which_osx_path = shutil.which('kindlegen')
            kg_path = which_osx_path

        # /Applications/calibre.app/Contents/MacOS/calibre-debug
        mac_calibre_debug_path = os.path.join('/Applications', 'calibre.app', 'Contents', 'MacOS', 'calibre-debug')
        if os.path.isfile(mac_calibre_debug_path):
            calibre_debug_path = mac_calibre_debug_path
        else:
            which_osx_path = shutil.which('kindlegen')
            if which_osx_path: calibre_debug_path = which_osx_path

    # display select file dialog box
    if not kg_path:
        if iswindows: 
            _filter = ['kindlegen.exe (*.exe)', 'All files (*.*)']
        else:
            _filter = ['kindlegen (kindlegen)', 'All files (*.*)']
        kg_path = select_file(bk, 'Select kindlegen executable', _filter)
        if not kg_path or not os.path.basename(kg_path).startswith('kindlegen') or kg_path.endswith('.zip'):
            kg_path = None
    return kg_path, calibre_debug_path

#------------------------------------------
# look for CFF/Type 1 (PostScript) fonts
#------------------------------------------
def cff_fonts(bk, ebook_root):
    ''' returns a list of CFF/Type 1 fonts '''
    # look for cff fonts
    font_manifest_items = []
    font_manifest_items = list(bk.font_iter())
    cff_fonts = []

    # fonts aren't found by bk.font_iter(), if they have the wrong mime type, e.g. 'application/octet-stream
    if font_manifest_items == []:
        for manifest_id, href, mime in bk.manifest_iter():
            if href.endswith('tf'):
                font_manifest_items.append((manifest_id, href, mime))

    for manifest_id, href, mime in font_manifest_items:
        if  bk.launcher_version() >= 20190927:
            font_path = os.path.join(ebook_root, bk.id_to_bookpath(manifest_id))
        else:
            font_path = os.path.join(ebook_root, 'OEBPS', href)
        with open(font_path, 'rb') as f:
            # CFF/Type 1 fonts start with b'OTTO'
            magic_number = f.read(4)
            if magic_number == b'OTTO':
                cff_fonts.append(os.path.basename(href))

    return cff_fonts

#-------------------------------
# get cover image information
#------------------------------
def get_cover_info(bk, cover_id):
    ''' returns cover image width, height, xdpi and ydpi '''
    width = height = xdpi = ydpi = None
    if cover_id and bk.id_to_href(cover_id): 
        if os.path.splitext(bk.id_to_href(cover_id))[1][1:].upper() != 'SVG':
            imgdata = bk.readfile(cover_id)
            try:
                img = Image.open(BytesIO(imgdata)).convert('L')
                height, width = img.size # width should be at least 500
                xdpi, ydpi = img.info['dpi'] # should be at least 300dpi
            except Exception as ex:
                if type(ex).__name__ != 'KeyError': 
                    print('*** PYTHON ERROR ***\nAn exception of type {0} occurred.\nArguments:\n{1!r}'.format(type(ex).__name__, ex.args))
                pass
    return height, width, xdpi, ydpi

#---------------------------
# generate thumbnail
#---------------------------
def generate_thumbnail(bk, cover_id, thumbnail_height, img_dest_path):
    ''' generates a kindle thumbnail file '''
    if cover_id and bk.id_to_href(cover_id): 
        if os.path.splitext(bk.id_to_href(cover_id))[1][1:].upper() != 'SVG':
            imgdata = bk.readfile(cover_id)
            try:
                img = Image.open(BytesIO(imgdata)).convert('L')
                # check for Pillow 10.x, which deprecateted Image.ANTIALIAS
                if hasattr(Image, '__version__') and int(Image.__version__.split('.')[0]) >= 10:
                    img.thumbnail((thumbnail_height, thumbnail_height), Image.Resampling.LANCZOS)
                else:
                    img.thumbnail((thumbnail_height, thumbnail_height), Image.ANTIALIAS)
                img.save(img_dest_path)
            except Exception as ex:
                if type(ex).__name__ != 'KeyError': 
                    print('*** PYTHON ERROR ***\nAn exception of type {0} occurred.\nArguments:\n{1!r}'.format(type(ex).__name__, ex.args))
                pass
    else:
        print('Cover image not found!')

#----------------------------
# simple kindlegen wrapper
#----------------------------
def kgWrapper(*args):
    '''simple KindleGen wrapper '''
    process = Popen(list(args), stdout=PIPE, stderr=PIPE)
    ret = process.communicate()
    return ret

#----------------------------
# reverts first last names
#----------------------------
def LastFirst(author):
    ''' reverses the name of the author '''
    author_parts = author.split(' ')
    if len(author_parts) >= 2:
        LastFirst = author_parts[len(author_parts) - 1] + ', ' + " ".join(author_parts[0:len(author_parts) - 1])
        return LastFirst
    else:
        return author

#-----------------------------------
# file selection dialog
#-----------------------------------
def select_file(bk, title, _filter):
    ''' simple file selection dialog box '''
    file_path = None
    if not PluginApplication.instance():
        app = PluginApplication(sys.argv, bk)
    else:
        app = PluginApplication.instance()
    dlg = QtWidgets.QFileDialog()
    if isosx:
        dlg.setOption(QtWidgets.QFileDialog.DontUseNativeDialog)
    dlg.setWindowTitle(title)
    dlg.setNameFilters(_filter)
    dlg.setFileMode(QtWidgets.QFileDialog.AnyFile)
    if dlg.exec():
        filenames = dlg.selectedFiles()
        file_path = filenames[0]
    return file_path

#------------------------------------------
# ok message box
#------------------------------------------
def ok_msg_box(title, text, bk):
    ''' displays a simple OK Cancel message box '''
    ret_val = False
    if not PluginApplication.instance():
        app = PluginApplication(sys.argv, bk)
    else:
        app = PluginApplication.instance()
    msg = QtWidgets.QMessageBox()
    msg.setWindowTitle(title)
    msg.setText(text)
    msg.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
    button_clicked = msg.exec()
    if button_clicked == QtWidgets.QMessageBox.Ok:
        ret_val = True
    return ret_val

#===================
# define GUI
#===================
class GUI(QtWidgets.QDialog):

    def __init__(self, bk, srl, toc, cover, prefs, asin):

        #==========================
        # initialize dialog
        #==========================

        # get variables
        self.bk = bk
        self.srl = srl
        self.toc = toc
        self.toc = toc
        self.asin = asin
        self.prefs = prefs
        self.onlyInt = QtGui.QIntValidator() # to enforce numerical input for thumbnails

        # initialize class
        super(GUI, self).__init__()

        # load gui file
        loadUi(os.path.join(bk._w.plugin_dir, bk._w.plugin_name, 'main.ui'), self)

        #======================
        # connections
        #======================

        # OK button
        self.buttonBox.accepted.connect(self.close)
        self.accepted.connect(self.save_values)

        # Cancel button
        # define connects (A “Close” button defined with the RejectRole)
        self.buttonBox.rejected.connect(self.close)
        self.rejected.connect(self.cancel_plugin)

        # disable KFX for AZW3/MOBI7 and vice versa
        self.chkGenKFX.stateChanged.connect(self.disable_mobi)
        self.chkGenAZW3.stateChanged.connect(self.disable_kfx)
        self.chkGenMOBI7.stateChanged.connect(self.disable_kfx)

        # generate ASIN, if necessary
        self.chkAddASIN.stateChanged.connect(self.check_asin)

        # update ASIN, if the text box was changed
        self.leASIN.textChanged.connect(self.update_asin)

        # folder selection dialog
        self.tbBrowse.clicked.connect(self.select_folder)

        #=====================================
        # initialize controls
        #=====================================

        #--------------------------
        # set metadata values
        #--------------------------
        if 'check_srl' in prefs and prefs['check_srl'] == False:
            self.lblSRL.setStyleSheet("color: blue;")
            self.lblSRL.setText('SRL: ignored')
        else:
            if srl:
                self.lblSRL.setText('SRL: ' + srl)
            else:
                self.lblSRL.setStyleSheet("color: red;")
                self.lblSRL.setText('SRL: not found')

        if toc:
            self.lblTOC.setText('TOC: ' + toc)
        else:
            self.lblTOC.setStyleSheet("color: red;")
            self.lblTOC.setText('TOC: not found')

        if cover:
            self.lblCover.setText('Cover: ' + cover)
        else:
            self.lblCover.setStyleSheet("color: red;")
            self.lblCover.setText('Cover: not found')

        if asin:
            self.leASIN.setText(asin)
        else:
            if 'add_asin' in prefs and prefs['add_asin'] == True:
                asin = str(uuid.uuid4())[24:82]
                self.asin = asin
                self.leASIN.setText(asin)

        #-----------------------
        # set check boxes
        #-----------------------
        if 'donotaddsource' in prefs:
            self.chkDoNotAddSource.setChecked(prefs['donotaddsource'])
        if 'verbose' in prefs:
            self.chkVerbose.setChecked(prefs['verbose'])
        if 'western' in prefs:
            self.chkWestern.setChecked(prefs['western'])
        if 'gif' in prefs:
            self.chkGif.setChecked(prefs['gif'])
        if 'add_asin' in prefs:
            self.chkAddASIN.setChecked(prefs['add_asin'])
        if 'thumbnail' in prefs:
            self.chkGenThumb.setChecked(prefs['thumbnail'])
        if 'azw3_only' in prefs:
            self.chkGenAZW3.setChecked(prefs['azw3_only'])
        if islinux: self.chkGenKFX.setEnabled(False)
        if 'kfx' in prefs and not islinux:
            self.chkGenKFX.setChecked(prefs['kfx'])
        if 'mobi7' in prefs:
            self.chkGenMOBI7.setChecked(prefs['mobi7'])

        #--------------------
        # set combo boxes
        #--------------------
        self.cbCompression.addItems(['0', '1', '2'])
        if 'compression' in prefs:
            compression = int(prefs['compression'])
            if compression not in [0, 1, 2] : compression = 1
        else:
            compression = 1
        self.cbCompression.setCurrentIndex(compression)

        locales = ['en', 'de', 'fr', 'it', 'es', 'zh', 'ja', 'pt', 'ru']
        self.cbLocale.addItems(locales)
        if 'locale' in prefs:
            locale = prefs['locale']
            if locale not in locales: locale = 'en'
        else:
            locale = 'en'
        index = [i for i, x in enumerate(locales) if x == locale]
        self.cbLocale.setCurrentIndex(int(index[0]))

        #-------------------
        # set text boxes
        #-------------------
        if 'mobi_dir' in prefs:
            mobi_dir = prefs['mobi_dir']
            if not os.path.isdir(mobi_dir): mobi_dir = GetDesktop()
        else:
            mobi_dir = GetDesktop()
        self.leOutDir.setText(mobi_dir)

        # ensure that only numbers can be entered
        self.leHeight.setValidator(self.onlyInt)
        if 'thumbnail_height' in prefs:
            thumbnail_height = prefs['thumbnail_height']
        else:
            thumbnail_height = 330
        self.leHeight.setText(str(thumbnail_height))

        #------------------------
        # show dialog box
        #------------------------
        self.show()

    #===============================
    # save values
    #===============================
    def save_values(self):
        ''' saves the preference options '''

        #check boxes
        self.prefs['donotaddsource'] = self.chkDoNotAddSource.isChecked()
        self.prefs['verbose'] = self.chkVerbose.isChecked()
        self.prefs['western'] = self.chkWestern.isChecked()
        self.prefs['gif'] = self.chkGif.isChecked()
        self.prefs['add_asin'] = self.chkAddASIN.isChecked()
        if not self.prefs['add_asin']: self.asin = None
        self.prefs['thumbnail'] = self.chkGenThumb.isChecked()
        self.prefs['azw3_only'] = self.chkGenAZW3.isChecked()
        self.prefs['mobi7'] = self.chkGenMOBI7.isChecked()
        self.prefs['kfx'] = self.chkGenKFX.isChecked()

        #combo boxes
        self.prefs['locale'] = self.cbLocale.currentText()
        self.prefs['compression'] = int(self.cbCompression.currentText())

        #text boxes
        self.prefs['mobi_dir'] = self.leOutDir.text().strip()
        self.prefs['thumbnail_height'] = int(self.leHeight.text().strip())

        self.bk.savePrefs(self.prefs)

        global asin
        asin = self.asin

        # set cancel status variable
        global cancel
        cancel = False

    #====================
    # cancel plugin button
    #====================
    def cancel_plugin(self):
        # set cancel status variable
        global cancel
        cancel = True

    #-------------------------------
    # auxiliary functions
    #--------------------------------

    def disable_mobi(self):
        ''' disable azw3/mobi7 options if kfx was selected '''
        if self.chkGenKFX.isChecked():
            self.chkGenAZW3.setChecked(False)
            self.chkGenAZW3.setEnabled(False)
            self.chkGenMOBI7.setChecked(False)
            self.chkGenMOBI7.setEnabled(False)
        else:
            self.chkGenAZW3.setEnabled(True)
            self.chkGenMOBI7.setEnabled(True)

    def disable_kfx(self):
        ''' disable kfx option if azw3/mobi7 output was selected on Linux machines'''
        if not islinux:
            if self.chkGenAZW3.isChecked() or self.chkGenMOBI7.isChecked():
                self.chkGenKFX.setChecked(False)
                self.chkGenKFX.setEnabled(False)
            else:
                self.chkGenKFX.setEnabled(True)

    def check_asin(self):
        ''' generate a fake ASIN '''
        if self.chkAddASIN.isChecked(): 
            if self.asin:
                self.leASIN.setText(self.asin)
            else:
                self.asin = str(uuid.uuid4())[24:82]
                self.leASIN.setText(self.asin)
        else:
            self.asin = None
            self.leASIN.setText('')

    def update_asin(self):
        ''' updates the asin value passed on to the main routine '''
        self.asin = self.leASIN.text()

    def select_folder(self):
        ''' simple folder selection dialog box '''
        mobi_dir = str(QtWidgets.QFileDialog().getExistingDirectory(self, "Select output directory"))
        if (os.path.isdir(mobi_dir)): 
            self.leOutDir.setText(mobi_dir)
        else:
            self.leOutDir.setText(GetDesktop())


#=============================
# main plugin routine
#=============================
def run(bk):
    ''' the main routine '''
    # PyQt/PySide requires at least Sigil 0.9.8
    if not bk.launcher_version() >= 20170115:
        print('This plugin requires Sigil 0.9.8 or higher.\nClick OK to close the Plugin Runner window.')
        return 0

    #----------------------------
    # get preferences
    #----------------------------
    debug = False
    kg_path = None
    calibre_debug_path = None
    prefs = bk.getPrefs()


    # save default preferences
    if prefs == {}:
        kg_path, calibre_debug_path = findKindleGen(bk)
        print('Kindlegen path: {}\ncalibre-debug path: {}\n'.format(kg_path, calibre_debug_path))

        if not kg_path or not os.path.isfile(kg_path):
            print('kindlegen not found!\nPlease click OK to close the Plugin Runner window.')
            return -1
        else:
            # plugin options
            prefs['kg_path'] = kg_path
            prefs['calibre_debug_path'] = calibre_debug_path
            prefs['quality_check'] = True
            prefs['check_srl'] = True
        bk.savePrefs(prefs)

    # make sure that the kindlegen path is correct
    else:
        kg_path = prefs.get('kg_path', False)
        calibre_debug_path = prefs.get('calibre_debug_path', False)
        debug = prefs.get('debug', False)
        if debug: print('*** DEBUG ***\nPreferences:\nKindlegen path: {}\ncalibre-debug path: {}\n'.format(kg_path, calibre_debug_path))
        if  not kg_path or not os.path.isfile(kg_path):
            kg_path, calibre_debug_path = findKindleGen(bk)
            if kg_path and os.path.isfile(kg_path): 
                prefs['kg_path'] = kg_path
                if debug: print('*** DEBUG ***\nkg_path updated:', kg_path)
                bk.savePrefs(prefs)

    # double-check one more time that the kindlegen path is correct
    if not kg_path or not os.path.isfile(kg_path):
        if debug: print('*** DEBUG ***\nKindlegen path: {}\ncalibre-debug path: {}\n'.format(kg_path, calibre_debug_path))
        print('kindlegen not found!\nPlease click OK to close the Plugin Runner window.')
        return -1

    #========================================================
    # get guide/landmark items and the cover image id 
    #========================================================
    global asin
    srl = toc = cover = None
    epubversion = bk.epub_version()
    python3lib = find_python3lib_dir(bk._w.appdir)
    if not python3lib:
        print('python3lib not found.\nPlease click OK to close the Plugin Runner window.')
        return -1
    else:
        # Prepend python3lib directory to sys.path
        if os.path.exists(python3lib) and os.path.isdir(python3lib):

            # import Opf_Parser()
            sys.path.insert(0, os.path.abspath(python3lib))
            from opf_newparser import Opf_Parser
            op = Opf_Parser(bk.get_opf())

            #-------------------------------------------
            # get title, author and epub2 cover_image_id
            #-------------------------------------------
            dtb_uid = cover_image_meta_id = cover_image_props_id = dc_language = _id = None
            asin = dc_asin = kfx_asin = None
            dc_creator = 'unknown'
            dc_title = 'untitled'
            for mname, mcontent, keylist, vallist in op.get_metadata():
                # ignore empty metadata entries
                if mcontent:
                    if mname == 'dc:title': 
                        if dc_title == 'untitled':
                            dc_title = mcontent
                        else:
                            dc_title += ' ' + mcontent
                    if mname == 'dc:creator' and dc_creator == 'unknown': 
                        dc_creator = LastFirst(mcontent)
                        dc_creator = re.sub(r'[/|\?|<|>|\\\\|:|\*|\||"|\^| ]+', '_', dc_creator)
                    if mname == 'dc:language': 
                        dc_language = mcontent
                    if mname == 'dc:identifier': 
                        dc_identifier = mcontent
                        if keylist == ['opf:scheme'] and vallist[0].upper() in ('AMAZON', 'MOBI-ASIN'): 
                            dc_asin = dc_identifier
                        if dc_identifier.startswith('urn'):
                            urn_list = dc_identifier.split(':')
                            if len(urn_list) == 3 and urn_list[1].upper() in ('AMAZON', 'MOBI-ASIN'):
                                dc_asin = urn_list[2]
                        if dc_identifier.startswith('amazon:'):
                            dc_asin = dc_identifier.split(':')[1]
                        if not asin and dc_asin: 
                            asin = dc_asin
                        else:
                            if dc_asin and dc_asin.startswith('B'):
                                asin = dc_asin
                else:
                    if debug: print('Empty metadata entry found:', mname, mcontent, keylist, vallist)

                if mname == 'meta':
                    if len(keylist) == 2 and len(vallist) == 2:
                        if keylist[0] == 'name' and vallist[0] == 'cover':
                            _id = vallist[1]
                        if keylist[1] == 'name' and vallist[1] == 'cover':
                            _id = vallist[0]

                        if _id:
                            if not bk.id_to_href(_id):
                                print('\nThe EPUB2 cover image referenced by the id "{}" is not manifested! ⚠'.format(_id))
                            else:
                                cover_image_meta_id = _id
                                cover = os.path.basename(bk.id_to_href(cover_image_meta_id))
                                
            dc_title = re.sub(r'[/|\?|<|>|\\\\|:|\*|\||"|\^| ]+', '_', dc_title)

            if debug: print('cover_image_meta_id:', cover_image_meta_id)

            #============================================
            # kindlegen requires a language code
            #============================================
            if not dc_language:
                print('\nError: Language metadata entry missing!\n\nPlease click OK to close the Plugin Runner window.')
                return 0

            if debug: print('metadata', asin, dc_title, dc_creator, dc_language, cover_image_meta_id)

            #------------------------------
            # get guide items
            #------------------------------
            guide_types = []
            guide_items = op.get_guide()
            if guide_items != []:
                for guide_type, _, href in guide_items:
                    if guide_type not in guide_types:
                        if guide_type == 'text': srl = os.path.basename(href)
                        if guide_type == 'toc': toc = os.path.basename(href)
                        guide_types.append(guide_type)
            else:
                guide_types = []

            if debug: print('guide_types:', guide_types)

            if epubversion.startswith('3'):
                #-------------------------------------------
                # check for guide item cover property id
                #-------------------------------------------
                cover_image_props_id = None
                for _id, href, mtype, keylist, vallist in op.get_manifest():
                    if 'cover-image' in vallist:
                        if not bk.id_to_href(_id):
                            print('\nThe EPUB3 cover image referenced by the id "{}" not manifested! ⚠'.format(cover_image_props_id))
                        else:
                            cover_image_props_id = _id
                            cover = os.path.basename(bk.id_to_href(cover_image_props_id))
                        break
                if debug: print('cover_image_props_id:', cover_image_props_id)


                #================================
                # landmarks handling
                #================================

                if bk.launcher_version() >= 20190716:
                    #----------------------------------------------------------------------------------------------
                    # use faster NavProcessor bundled with sigil 0.9.1.6 and higher
                    #----------------------------------------------------------------------------------------------
                    if debug: print('Using external NavProcessor for nav parsing...')

                    #-------------------------------------------------------------
                    # get nav_id, nav_basename and nav_data
                    #-------------------------------------------------------------
                    nav_id = bk.getnavid()
                    if nav_id:
                        nav_basename = os.path.basename(bk.id_to_href(nav_id))
                        nav_data = bk.readfile(nav_id)

                        #------------------------------
                        # get landmarks
                        #------------------------------
                        from navprocessor import NavProcessor
                        np = NavProcessor(nav_data)
                        landmark_list = np.getLandmarks()
                        if debug: print('landmark_list:', landmark_list)
                    else:
                        print('NAV landmark not found! Can\'t continue.\nClick OK to close.')
                        return -1

                else:
                    if debug: print('Using embedded NavProcessor for nav doc parsing...')
                    #---------------------------------------------------------------
                    # get nav_id, nav_basename and nav_data
                    #---------------------------------------------------------------
                    nav_id = getnavid(bk)
                    if nav_id:
                        nav_basename = os.path.basename(bk.id_to_href(nav_id))
                        nav_data = bk.readfile(nav_id)

                        #------------------------------
                        # get landmarks
                        #------------------------------
                        landmark_list = getLandmarks(nav_data)
                        if debug: print('landmark_list:', landmark_list)

                    else:
                        print('NAV landmark not found! Can\'t continue.\nPlease click OK to close the Plugin Runner window.')
                        return -1

                # enumerate landmark_types
                landmark_types = []
                for landmark_type, href, _ in landmark_list:
                    if not landmark_type in landmark_types:
                        landmark_types.append(landmark_type)
                        if landmark_type == 'bodymatter': srl = os.path.basename(href)
                        if landmark_type == 'toc':
                            toc = os.path.basename(href)
                            # add missig nav file name
                            if toc.startswith('#'): toc = nav_basename + toc
                if debug: print('landmark_types:', landmark_types)

    print('===================')
    if dc_creator != 'unknown': print('Author:', dc_creator.replace('_', ' '))
    if dc_title != 'untitled' : print('Title:', dc_title.replace('_', ' '))
    if srl: print('SRL:', srl, '✅')
    if toc: print('TOC:', toc, '✅')
    if cover: 
        print('Cover:', cover, '✅')
        # if epubversion.startswith('3') and not cover_image_props_id and cover_image_meta_id:
            # print('You might want to mark {} as the cover image.'.format(cover))
    else:
        print('Cover: not found', '⚠')
    if asin: print('ASIN:', asin)
    print('===================')

    #==============================
    # check for quality issues
    #==============================
    quality_check = prefs.get('quality_check', True)
    check_srl = prefs.get('check_srl', True)
    if quality_check:

        #------------------------------------
        # check for type 1 Postscript fonts
        #------------------------------------
        ps_fonts = cff_fonts(bk, bk._w.ebook_root)
        if ps_fonts != []:
            for ps_font in ps_fonts:
                print('CFF/Type1 (Postscript) font:', ps_font, '⚠')

        #-------------------------------
        # get cover image width
        #-------------------------------
        if cover_image_meta_id or cover_image_props_id:
            if cover_image_meta_id:
                height, width, xdpi, ydpi = get_cover_info(bk, cover_image_meta_id)
            else:
                width, height, xdpi, ydpi = get_cover_info(bk, cover_image_props_id)
            if debug: print('width, height, xdpi, ydpi:', width, height, xdpi, ydpi)

            if width < 500:
                print('Cover image problem: cover image width less than 500px:', width, '⚠')
            # if xdpi and ydpi and (xdpi < 300 or ydpi < 300):
                # print('Cover image problem: resolution less than 300 dpi:', xdpi, 'x', ydpi, '⚠')
        else:
            print('Cover image problem: cover image not found! ⚠')

        #------------------------------------
        # check for landmark/guide problems
        #------------------------------------
        if epubversion.startswith('3'):
            if check_srl and not 'bodymatter' in landmark_types: print('EPUB3 landmark problem found: start reading location not defined. ⚠')
            if not 'toc' in landmark_types: print('EPUB3 landmark problem found: HTML table of contents not defined. ⚠')
        else:
            if check_srl and not 'text' in guide_types: print('EPUB2 guide item problem found: start reading location not defined. ⚠')
            if not 'toc' in guide_types: print('EPUB2 guide item problem found: HTML table of contents not defined. ⚠')

        #-------------------------------------
        # check for unsupported CSS selectors
        #-------------------------------------
        # for css_id, href in bk.css_iter():
            # css = bk.readfile(css_id)
            # cssutils.log.setLevel(logging.FATAL)
            # sheet = cssutils.parseString(css)
            # for rule in sheet:
                # if rule.type == rule.STYLE_RULE:
                    # # find properties that start with max:
                    # for property in rule.style:
                        # if property.name.startswith('max'):
                            # print('Unsupported CSS property "{}" found.'.format(property.name))


    #==========================================
    # display main GUI 
    #==========================================
    global cancel
    cancel = False

    mdp = True if iswindows else False
    app_icon = os.path.join(plugin_dir, 'app.ico')
    if not PluginApplication.instance():
        app = PluginApplication(sys.argv, bk, app_icon=app_icon, match_dark_palette=mdp, 
		load_qtbase_translations=False, load_qtplugin_translations=False)
    else:
        app = PluginApplication.instance()
    window = GUI(bk, srl, toc, cover, prefs, asin)
    app.exec()

    if debug: print('asin after:', asin)

    if cancel:
        print('\nPlease click OK to close the Plugin Runner window.')
        return 0

    #===========================================
    # retrieve preferences
    #===========================================
    #kg_path = prefs.get('kg_path', False)
    #calibre_debug_path = prefs.get('calibre_debug_path', False)
    add_asin = prefs.get('add_asin', False)
    azw3_only = prefs.get('azw3_only', False)
    mobi7 = prefs.get('mobi7', False)
    kfx = prefs.get('kfx', False)
    thumbnail = prefs.get('thumbnail', False)
    mobi_dir = prefs.get('mobi_dir', GetDesktop())
    thumbnail_height = prefs.get('thumbnail_height', 330)
    donotaddsource = prefs.get('donotaddsource', False)
    compression = prefs.get('compression', 1)
    verbose = prefs.get('verbose', False)
    western = prefs.get('western', False)
    gif = prefs.get('gif', False)
    locale = prefs.get('locale', 'en')
    if debug: print(prefs)

    # create temp folder
    temp_dir = tempfile.mkdtemp()

    # get get content.opf
    if bk.launcher_version() >= 20190927:
        opf_path = os.path.join(temp_dir, bk.get_opfbookpath())
    else:
        opf_path = os.path.join(temp_dir, 'OEBPS', 'content.opf')

    # get opf file name
    opf_name = os.path.basename(opf_path)

    # copy book files
    bk.copy_book_contents_to(temp_dir)

    #===========================
    # generate output
    #===========================
    base_name = dc_creator + '_' + dc_title

    #-----------------------
    # generate a kfx file
    #-----------------------
    if (kfx and calibre_debug_path) and not islinux:
        kfx_asin = None

        # create mimetype file
        with open(os.path.join(temp_dir, "mimetype"), "w") as mimetype:
            mimetype.write("application/epub+zip")

        # zip up the epub folder and save the epub in the plugin folder
        epub_path = os.path.join(bk._w.plugin_dir, bk._w.plugin_name, 'temp.epub')
        if os.path.isfile(epub_path):
            os.remove(str(epub_path))
        epub_zip_up_book_contents(temp_dir, epub_path)

        # assemble parameters
        if add_asin and asin:
            kfx_path = os.path.join(mobi_dir, base_name  + '_' + asin + '.kfx')
            args = [calibre_debug_path, '-r', 'KFX Output', '--', '-a', asin, '-p', '0', epub_path, kfx_path]
        else:
            kfx_path = os.path.join(mobi_dir, base_name + '.kfx')
            args = [calibre_debug_path, '-r', 'KFX Output', '--', '-p', '0', epub_path, kfx_path]

        # run Calibre KFX Output plugin
        print("\nRunning Calibre KFX Output plugin ... please wait\n")
        try:
            if debug: print('args:', args)
            result = kgWrapper(*args)
        except FileNotFoundError as not_found:
            print('calibre-debug not found!\nPlease click OK to close the Plugin Runner window.')
            return -1

        # print Calibre messages
        calibre_messages = result[0].decode('utf-8', 'ignore')
        print(calibre_messages)

        # get the ASIN number generated by the KFX Plugin
        kfx_asin_match = re.search('ASIN=([^,]+),', calibre_messages)
        if kfx_asin_match is not None:
            kfx_asin = kfx_asin_match.group(1)
            asin = kfx_asin

        # if the plugin didn't fail, there should be a .kfx file
        if os.path.isfile(kfx_path):
            print('KFX file copied to ' + kfx_path)
        else:
            print('KFX file coudn\'t be generated.')

        # delete temp epub from plugin folder
        os.remove(str(epub_path))

    else:
        #-----------------------------------
        # generate file with kindlegen
        #-----------------------------------
        
        if kfx and not calibre_debug_path:
            print('calibre-debug/Kindle Previewer not found; generating mobi file.')

        #------------------------------------------
        # define kindlegen command line parameters
        #------------------------------------------

        # define temporary mobi file name
        if bk.launcher_version() >= 20190927: 
            OEBPS = bk.get_startingdir(bk.get_opfbookpath())
        else:
            OEBPS = 'OEBPS'
        mobi_path = os.path.join(temp_dir, OEBPS, 'sigil.mobi')

        args = [kg_path, opf_path]

        if compression in ['0', '1', '2'] or compression in [0, 1, 2]:
            args.append('-c' + str(compression))

        if donotaddsource and not kfx:
            args.append('-dont_append_source')

        if verbose:
            args.append('-verbose')

        if western:
            args.append('-western')

        if gif:
            args.append('-gif')

        if locale in ['en', 'de', 'fr', 'it', 'es', 'zh', 'ja', 'pt', 'ru']:
            args.append('-locale')
            args.append(locale)

        args.append('-o')
        args.append('sigil.mobi')

        # run kindlegen
        print("\nRunning KindleGen ... please wait")
        if debug: print('args:', args)
        result = kgWrapper(*args)

        # print kindlegen messages
        kg_messages = result[0].decode('utf-8', 'ignore').replace(temp_dir, '').replace('\r\n', '\n')
        print(kg_messages)

        # display kindlegen warning messages
        if re.search(r'(W|E)\d+:', kg_messages):
            print('\n*************************************************************\nkindlegen warnings:\n*************************************************************')
            kg_warnings = kg_messages.splitlines()
            for line in kg_warnings:
                if re.search(r'(W|E)\d+:', line) or line.startswith("Info(prcgen):I1052"):
                    print(line)
            print('\n*************************************************************\n')

        if not os.path.isfile(mobi_path):
            print('Kindlegen failed.\nPlease click OK to close the Plugin Runner window.')
            return -1

        # add asin number
        if add_asin and asin:
            dmf = DualMobiMetaFix(mobi_path, asin)
            open(pathof(mobi_path + '.tmp'), 'wb').write(dmf.getresult())

            # if DualMobiMetaFix didn't fail, there should be a .tmp file
            if os.path.isfile(str(mobi_path) + '.tmp'):
                print('\nASIN: ' + asin + ' added')

                # delete original file and rename temp file
                os.remove(str(mobi_path))
                os.rename(str(mobi_path) + '.tmp', str(mobi_path))
            else:
                print('\nASIN couldn\'t be added.')
                add_asin = False

        # write dual mobi file
        if not mobi7 and not azw3_only:
            if add_asin and asin:
                dst_path = os.path.join(mobi_dir, base_name + '_' + asin + '_dual.mobi')
            else:
                dst_path = os.path.join(mobi_dir, base_name + '_dual.mobi')

            shutil.copyfile(mobi_path, dst_path)
            if os.path.isfile(dst_path):
                print('\nDual mobi copied to: ' + dst_path)
            else:
                print('\nDual mobi NOT copied to: ' + dst_path, '\nPlease click OK to close the Plugin Runner window.')
                return -1

        # generate MOBI7 and AZW3 files
        else:
            # define output file names
            if add_asin and asin:
                outmobi8 = os.path.join(mobi_dir, base_name + '_' + asin + '.azw3')
                outmobi7 = os.path.join(mobi_dir, base_name + '_' + asin + '.mobi')
            else:
                outmobi8 = os.path.join(mobi_dir, base_name + '.azw3')
                outmobi7 = os.path.join(mobi_dir, base_name + '.mobi')

            # split the mobi file
            mobisplit = mobi_split(pathof(mobi_path))

            # check if mobisplit was successful
            if mobisplit.combo:
                if azw3_only:
                    open(outmobi8, 'wb').write(mobisplit.getResult8())
                    if os.path.isfile(outmobi8): print('AZW3 file copied to ' + outmobi8)
                if mobi7:
                    open(outmobi7, 'wb').write(mobisplit.getResult7())
                    if os.path.isfile(outmobi7): print('MOBI7 file copied to ' + outmobi7)
            else:
                print('\nPlugin Error: Invalid mobi file format.\nPlease click OK to close the Plugin Runner window.')
                return -1

    # delete temp folder
    shutil.rmtree(temp_dir)

    #------------------------
    # generate thumbnail
    #------------------------
    if thumbnail and (cover_image_meta_id or cover_image_props_id):
        if cover_image_props_id and not cover_image_meta_id: cover_image_meta_id = cover_image_props_id
        if add_asin or kfx_asin:
            img_dest_path = os.path.join(mobi_dir, 'thumbnail_' + asin + '_EBOK_portrait.jpg')
            generate_thumbnail(bk, cover_image_meta_id, thumbnail_height, img_dest_path)

            # check if the thumbnail was actually copied
            if os.path.isfile(img_dest_path):
                print('\nThumbnail copied to: ' + img_dest_path)
            else:
                print('\nPlugin Error: Thumbnail creation failed.')

    print('\nPlease click OK to close the Plugin Runner window.')

    return 0

def main():
    print('I reached main when I should not have\n')
    return -1

if __name__ == "__main__":
    sys.exit(main())
